/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/
#define FILECODE "F(301)"

/*
 * vmx86.c --
 *
 *     Platform independent routines for creating/destroying/running
 *     virtual machine monitors.
 *
 */

#ifdef linux
/* Must come before any kernel header file --hpreg */
#   include "driver-config.h"

#   include <linux/string.h> /* memset() in the kernel */
#   include <linux/sched.h> /* jiffies from the kernel */
#elif defined(WINNT_DDK)
#   include <string.h>
#else
#   error "Unknown platform"
#endif

#include "vmware.h"
#include "machine.h"
#include "modulecall.h"
#include "vmx86.h"
#include "task.h"
#include "initblock.h"
#include "vm_asm.h"
#include "iocontrols.h"
#ifdef USE_PERFCTRS_HOSTED
#   include "perfctr.h"
#endif
#include "hostif.h"
#include "memtrack.h"


/*
 * Keep track of the virtual machines that have been
 * created using the following structures.
 *
 */

static VMDriver *vmDriverList = NULL;

static Bool  lockedPageLimitMonitor;

/* Max pages that all virtual machines can cumulatively lock */
static uintptr_t lockedPageLimit;

/* 
 * Max pages that all virtual machines can cumulatively lock according to
 * vmmon's estimates. This limit is always enforced, otherwise the host
 * would die or be really unhappy. 
 */
static uintptr_t unconditionalLockedPageLimit;

/* Number of pages actually locked by all virtual machines */
static uintptr_t numLockedPages;

/* Total minAllocation of all vms on this host */
static uintptr_t globalMinAllocation;

/* Total virtual machines on this host */
static uintptr_t vmCount;

static void FreeAllResources(VMDriver *vm);


/*
 *----------------------------------------------------------------------
 *
 * Vmx86HasFreePages --
 *
 *       Returns TRUE if the vm can lock more pages.  This is true if 
 *       we are below the host's hard memory limit and this vm has not
 *       exceeded its maximum allocation.
 *       Callers must ensure ensure driver-wide and VM serialization
 *       typically by using HostIF_GlobalLock() and  HostIF_VMLock().
 *
 * Results:
 *       TRUE if pages can be locked, FALSE otherwise
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

static INLINE Bool
Vmx86HasFreePages(VMDriver *vm, 
		  unsigned int numPages)
{
   /*  
    * 1) Be careful with overflow. 
    * 2) lockedPageLimit and vm->memInfo.maxAllocation can be decreased below
    *    the current numLockedPages and vm->memInfo.locked
    * 3) unconditionalLockedPageLimit can NOT go lower than numLockedPages,
    *    unconditionalLockedPageLimit should be set only once on driver loading.
    */

   ASSERT(unconditionalLockedPageLimit >= numLockedPages);

   if (numPages > unconditionalLockedPageLimit - numLockedPages ) {
      return FALSE;
   }
   
   if (lockedPageLimit != 0) {
      if (lockedPageLimit <= numLockedPages) {
	 return FALSE;
      } else if (lockedPageLimit - numLockedPages < numPages) {
	 return FALSE;
      }
   }

   if (vm->memInfo.valid) {
      if (vm->memInfo.maxAllocation <= vm->memInfo.locked) {
	 return FALSE;
      } else if (vm->memInfo.maxAllocation - vm->memInfo.locked < numPages) {
	 return FALSE;
      }
   }

   return TRUE;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86ReserveFreePages --
 *
 *       Returns TRUE and increases locked page counts if the vm can lock 
 *       more pages.  This is true if we are below the host's hard memory 
 *       limit and this vm has not exceeded its maximum allocation.
 *       The function is thread-safe.
 *
 * Results:
 *       TRUE if pages are reserved for locking, FALSE otherwise
 *
 * Side effects:
 *       The global lock and VM's lock are acquired and released.
 *
 *----------------------------------------------------------------------
 */

static Bool
Vmx86ReserveFreePages(VMDriver *vm, 
		      unsigned int numPages)
{
   Bool retval;

   ASSERT(vm);

   HostIF_GlobalLock(17);
   HostIF_VMLock(vm);

   retval = Vmx86HasFreePages(vm, numPages);
   if (retval) {
      numLockedPages += numPages;
      vm->memInfo.locked += numPages;
      vm->stats.monitorLockedPages += numPages;
      if (vm->stats.maxMonitorLockedPages > vm->stats.monitorLockedPages - vm->stats.userLockedPages) {
         vm->stats.maxMonitorLockedPages = vm->stats.monitorLockedPages - vm->stats.userLockedPages;
      }
   }
   HostIF_VMUnlock(vm);
   HostIF_GlobalUnlock(17);

   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86UnreserveFreePages --
 *
 *       Decreases the global and VM's locked page counts. 
 *       The function is thread-safe. 
 *
 * Results:
 *       void
 *
 * Side effects:
 *       The global lock and VM's lock are acquired and released.
 *
 *----------------------------------------------------------------------
 */

static void
Vmx86UnreserveFreePages(VMDriver *vm, 
			unsigned int numPages)
{
   ASSERT(vm);

   HostIF_GlobalLock(18);
   HostIF_VMLock(vm);

   ASSERT(numLockedPages >= numPages);
   ASSERT(vm->memInfo.locked >= numPages);

   numLockedPages -= numPages;
   vm->memInfo.locked -= numPages;
   vm->stats.monitorLockedPages -= numPages;

   HostIF_VMUnlock(vm);
   HostIF_GlobalUnlock(18);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_CreateVM --
 *
 *      Allocate and initialize a driver structure for a virtual machine.
 *
 * Results:
 *      VMDriver structure or NULL on error.
 *
 * Side effects:
 *       Memory allocated.
 *
 *----------------------------------------------------------------------
 */

VMDriver *
Vmx86_CreateVM(void *uniqHandle,             // IN: Unique id for VM to be
                                             //     created
	       uintptr_t processId,          // IN: Process creating VM
	       WSLimitsInfo *wsLimitsInfoIn, // IN: Information about resource
                                             //     limits
	       enum vmEmulations vmVersion)
{
   VMDriver *vm;
   VMDriver **vmp;

   vm = HostIF_AllocKernelMem(sizeof *vm, TRUE);
   if (vm == NULL) {
      return NULL;
   }
   memset(vm, 0, sizeof *vm);
#ifdef linux
   vm->logFD = -1;
#endif

   ASSERT(sizeof (uintptr_t) == sizeof vm);
   vm->id = ((uintptr_t)vm & 0xffffffff) >> 1;
   vm->uniqHandle = uniqHandle;
   vm->processID = processId;
   vm->memInfo.valid = FALSE;

   memcpy(&vm->wsLimitsInfo, (char *)wsLimitsInfoIn, sizeof vm->wsLimitsInfo);

   if (HostIF_Init(vm)) {
      goto cleanup;
   }

   HostIF_GlobalLock(0);

   if (vmCount >= MAX_VMS) {
      HostIF_GlobalUnlock(0);
      goto cleanup;
   }
   vmCount++;

   vm->count++;
   for (vmp = &vmDriverList; *vmp != NULL; vmp = &(*vmp)->nextDriver) {
   }
   *vmp = vm;

   switch (vmVersion) {
      case VME_TOT:
         vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_TOT)+ offsetof(VMCrossPageTOT,monContext));
	 break;
      case VME_V4:
	 vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_V4)+ offsetof(VMCrossPageV4,monContext));
	 break;
      case VME_GSX251:
      case VME_GSX25:
	 vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_GSX25)+ offsetof(VMCrossPageGSX25,monContext));
	 break;
      case VME_GSX2:
	 vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_GSX2)+ offsetof(VMCrossPageGSX2,monContext));
	 break;
      case VME_V321:
	 vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_V321)+ offsetof(VMCrossPageV321,monContext));
	 break;
      case VME_V32:
	 vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_V32)+ offsetof(VMCrossPageV32,monContext));
	 break;
      case VME_V3:
	 vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_V3)+ offsetof(VMCrossPageV3,monContext));
	 break;
      case VME_V2:
	 vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_V2)+ offsetof(VMCrossPageV3,monContext));
	 break;
      case VME_GSX1:
	 vm->monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START_GSX1)+ offsetof(VMCrossPageV3,monContext));
	 break;
      default:
         HostIF_GlobalUnlock(0);
	 Vmx86_ReleaseVM(vm);
	 printk(KERN_ERR "Unknown emulation %u\n", vmVersion);
	 return NULL;
   }
   vm->vmVersion = vmVersion;

   HostIF_GlobalUnlock(0);

   return vm;

cleanup:
   FreeAllResources(vm);
   return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_BindVM  --
 *
 *      Bind to an existing VM.
 *
 * Results:
 *      VMDriver structure or NULL on error.
 *
 * Side effects:
 *	Increment VM reference count.
 *
 *----------------------------------------------------------------------
 */

VMDriver *
Vmx86_BindVM(int id)
{
   VMDriver *vm;

   HostIF_GlobalLock(14);
   for (vm = vmDriverList; vm != NULL; vm = vm->nextDriver) {
      if (vm->id == id) {
	 vm->count++;
	 break;
      }
   }
   HostIF_GlobalUnlock(14);
   return vm;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_ReleaseVM  --
 *
 *      Release a VM (either created here or from a bind).
 *
 * Results:
 *      zero if successful
 *
 * Side effects:
 *	Decrement VM reference count.
 *      Release resources (those that are left) when count reaches 0.
 *
 *----------------------------------------------------------------------
 */
int
Vmx86_ReleaseVM(VMDriver *vm)
{
   VMDriver **vmp;

   HostIF_GlobalLock(1);

   /*
    * Do reference counting first
    */

   if (--vm->count > 0) {
      HostIF_GlobalUnlock(1);
      return 0;
   }

   ASSERT(vm->count == 0);

   /*
    *   Deletion of the virtual machine from list
    */

   for (vmp = &vmDriverList; *vmp != vm; vmp = &(*vmp)->nextDriver) {
      ASSERT(*vmp != NULL);
   }
   *vmp = vm->nextDriver;

   vmCount--;

   HostIF_GlobalUnlock(1);

   FreeAllResources(vm);

   return 0;
}

/*
 *------------------------------------------------------------------------------
 *
 * Vmx86_InitVM --
 *
 *    Initializaiton of the VM.  Expects all initial arguments
 *    to be part of the InitBlock structure.
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    Many
 *
 *------------------------------------------------------------------------------
 */

int
Vmx86_InitVM(VMDriver *vm,          // IN
             InitBlock *initParams) // IN: Initial params from the VM
{
   int retval;

   if (initParams->magicNumber != INIT_BLOCK_MAGIC) {
      Warning("Bad magic number for init block 0x%x\n", initParams->magicNumber);
      return 1;
   }
   if (initParams->numVCPUs >= MAX_INITBLOCK_CPUS) {
      Warning("Too many VCPUs for init block %d\n", initParams->numVCPUs);
      return 1;
   }
   vm->numVCPUs = initParams->numVCPUs;
   vm->logFD    = initParams->logFD;

   HostIF_InitFP(vm);
   HostIF_InitEvent(vm);

   /*
    * Initialize the driver's part of the cross-over page used to
    * talk to the monitor
    */

   retval = Task_InitCrosspage(vm, initParams);
   if (retval) {
      Warning("Task crosspage init died with retval=%d\n", retval);
      goto error;
   }

   return 0;

error:
   FreeAllResources(vm);
   return 1;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_LateInitVM --
 *
 *      Do late initialization of the driver.
 *	This should be called after Vmx86_CreateVM and
 *	after all the user-level device initialization.
 *
 * Results: 
 *	non-zero on error, zero on success;
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_LateInitVM(VMDriver *vm)
{
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_RunVM_TOT  --
 *
 *      Main interaction between the module and the monitor:
 *
 *	   Run the monitor
 *	   Process module calls from the monitor
 *	   Make cross user calls to the main thread
 *	   Return to userlevel to process normal user calls
 *	   and to signal timeout or errors.
 *
 * Results:
 *      Positive: user call number.
 *	-1: error (currently not used).
 *
 * Side effects:
 *      Not really.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_RunVM_TOT(VMDriver *vm, Vcpuid vcpuid)
{
   uint32 retval = MODULECALL_TOT_USERRETURN;
   VMCrossPageTOT *crosspage = vm->crosspage[vcpuid];

   ASSERT(crosspage);

   /*
    * Check if we were interrupted by signal.
    */
   if (crosspage->moduleCallInterrupted) {
      crosspage->moduleCallInterrupted = FALSE;
      goto skipTaskSwitch;
   }
   for (;;) {

      crosspage->retval = retval;

      /*
       * Task_Switch changes the world to the monitor.
       * The monitor is waiting in the BackToHost routine.
       */
      UCTIMESTAMP(crosspage, SWITCHING_TO_MONITOR);
      Task_Switch_TOT(vm, vcpuid);
      UCTIMESTAMP(crosspage, SWITCHED_TO_MODULE);

skipTaskSwitch:;

      retval = MODULECALL_TOT_USERRETURN;

      if (crosspage->userCallType != MODULECALL_USERCALL_NONE) {
         /*
          * This is the main user call path.
          *
          * There are two kinds of user calls.  Normal ones
          * are handled by the calling VCPU thread itself.
          * We just return from here (back to userlevel)
          * in that case.
          *
          * Calls marked userCallCross are handled by the main
          * VMX thread.  In this case, the userCallRequest field
          * indicates to the VMX that this VCPU wants to make
          * a user call.  This field may be consulted by the VMX
          * at any time (specifically when the VMX is awakened by
          * another VCPU), so it must be set after the other
          * user call arguments.  The VMX is responsible for
          * resetting this field and awakening the VCPU when
          * the user call is complete, via the ACK_USER_CALL
          * and COMPLETE_USER_CALL ioctl.  The latter implies
          * the former.
          *
          * When and how to use ACK_USER_CALL and COMPLETE_USER_CALL
          * are at the discretion of the VMX.  In particular,
          * COMPLETE_USER_CALL does not imply that the requested
          * operation has fully completed, only that the VCPU can
          * continue.  See the comment in MonitorLoopHandleUserCall()
          * for use scenarios.
          *
          * See also comment at MonitorLoopUserCallPoll().
          *
          * -- edward
          */

         if (!crosspage->userCallCross) {
            ASSERT(!crosspage->userCallRestart);
            return crosspage->userCallType;
         }

	 if (!crosspage->userCallRestart) {
	    ASSERT(crosspage->userCallRequest == MODULECALL_USERCALL_NONE);
	    crosspage->userCallRequest = crosspage->userCallType;
	    UCTIMESTAMP(crosspage, AWAKENING_VMX);
	    HostIF_UserCall(vm, vcpuid);
	 }

	 UCTIMESTAMP(crosspage, GOING_TO_SLEEP);
         if (HostIF_UserCallWait(vm, vcpuid)) {
	    ASSERT(crosspage->userCallRequest == MODULECALL_USERCALL_NONE);
	 } else {
	    retval = MODULECALL_TOT_USERTIMEOUT;
	 }
	 UCTIMESTAMP(crosspage, AWAKE);
      }

      switch (crosspage->moduleCallType) {

      case MODULECALL_TOT_NONE:
         break;

      case MODULECALL_TOT_INTR:    // Already done in task.c
         break;

      case MODULECALL_TOT_GET_RECYCLED_PAGE: {
	 MPN32 mpn;

	 retval = (Vmx86_AllocLockedPages(vm, &mpn, 1, TRUE) == 1) ?
	    mpn : INVALID_MPN;
	 break;
      }

      case MODULECALL_TOT_SEMAWAIT:
         retval = HostIF_SemaphoreWait(vm, vcpuid, crosspage->args[0]);
	 if (retval == MX_WAITINTERRUPTED) {
	    crosspage->moduleCallInterrupted = TRUE;
	    return USERCALL_RESTART;
	 }
         break;
      case MODULECALL_TOT_SEMASIGNAL:
         retval = HostIF_SemaphoreSignal(crosspage->args[0]);
	 if (retval == MX_WAITINTERRUPTED) {
	    crosspage->moduleCallInterrupted = TRUE;
	    return USERCALL_RESTART;
	 }
         break;
      case MODULECALL_TOT_SEMAFORCEWAKEUP: {
         HostIF_SemaphoreForceWakeup(vm, (Vcpuid) crosspage->args[0]);
         break;
      }
      case MODULECALL_TOT_IPI:
         retval = HostIF_IPI(vm, (VCPUSet)(uintptr_t)crosspage->args[0]);
         break;

      case MODULECALL_TOT_RELEASE_ANON_PAGE: {
         MPN32 mpn = (MPN32)crosspage->args[0];
	 retval = Vmx86_FreeLockedPages(vm, &mpn, 1, TRUE);
         break;
      }

      default:
         Warning("ModuleCall %d not supported\n", crosspage->moduleCallType);
      }

      /*
       * Give other threads and processes a chance.
       *
       * This is important when we are not preemptable while
       * in the driver and the monitor (e.g., Linux).
       */

      HostIF_Reschedule();
   }

   NOT_REACHED();
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_RunVM_V4  --
 *
 *      Main interaction between the module and the monitor:
 *
 *	   Run the monitor
 *	   Process module calls from the monitor
 *	   Make cross user calls to the main thread
 *	   Return to userlevel to process normal user calls
 *	   and to signal timeout or errors.
 *
 * Results:
 *      Positive: user call number.
 *	-1: error (currently not used).
 *
 * Side effects:
 *      Not really.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_RunVM_V4(VMDriver *vm, Vcpuid vcpuid)
{
   uint32 retval = MODULECALL_V4_USERRETURN;
   VMCrossPageV4 *crosspage = vm->crosspage[vcpuid];

   ASSERT(crosspage);

   /*
    * Check if we were interrupted by signal.
    */
   if (crosspage->moduleCallInterrupted) {
      crosspage->moduleCallInterrupted = FALSE;
      goto skipTaskSwitch;
   }
   for (;;) {

      crosspage->retval = retval;

      /*
       * Task_Switch changes the world to the monitor.
       * The monitor is waiting in the BackToHost routine.
       */
      UCTIMESTAMP(crosspage, SWITCHING_TO_MONITOR);
      Task_Switch_V4(vm, vcpuid);
      UCTIMESTAMP(crosspage, SWITCHED_TO_MODULE);

skipTaskSwitch:;

      retval = MODULECALL_V4_USERRETURN;

      if (crosspage->userCallType != MODULECALL_USERCALL_NONE) {
         /*
          * This is the main user call path.
          *
          * There are two kinds of user calls.  Normal ones
          * are handled by the calling VCPU thread itself.
          * We just return from here (back to userlevel)
          * in that case.
          *
          * Calls marked userCallCross are handled by the main
          * VMX thread.  In this case, the userCallRequest field
          * indicates to the VMX that this VCPU wants to make
          * a user call.  This field may be consulted by the VMX
          * at any time (specifically when the VMX is awakened by
          * another VCPU), so it must be set after the other
          * user call arguments.  The VMX is responsible for
          * resetting this field and awakening the VCPU when
          * the user call is complete, via the ACK_USER_CALL
          * and COMPLETE_USER_CALL ioctl.  The latter implies
          * the former.
          *
          * When and how to use ACK_USER_CALL and COMPLETE_USER_CALL
          * are at the discretion of the VMX.  In particular,
          * COMPLETE_USER_CALL does not imply that the requested
          * operation has fully completed, only that the VCPU can
          * continue.  See the comment in MonitorLoopHandleUserCall()
          * for use scenarios.
          *
          * See also comment at MonitorLoopUserCallPoll().
          *
          * -- edward
          */

         if (!crosspage->userCallCross) {
            ASSERT(!crosspage->userCallRestart);
            return crosspage->userCallType;
         }

	 if (!crosspage->userCallRestart) {
	    ASSERT(crosspage->userCallRequest == MODULECALL_USERCALL_NONE);
	    crosspage->userCallRequest = crosspage->userCallType;
	    UCTIMESTAMP(crosspage, AWAKENING_VMX);
	    HostIF_UserCall(vm, vcpuid);
	 }

	 UCTIMESTAMP(crosspage, GOING_TO_SLEEP);
         if (HostIF_UserCallWait(vm, vcpuid)) {
	    ASSERT(crosspage->userCallRequest == MODULECALL_USERCALL_NONE);
	 } else {
	    retval = MODULECALL_V4_USERTIMEOUT;
	 }
	 UCTIMESTAMP(crosspage, AWAKE);
      }

      switch (crosspage->moduleCallType) {

      case MODULECALL_V4_NONE:
         break;

      case MODULECALL_V4_INTR:    // Already done in task.c
         break;

      case MODULECALL_V4_ISMPNLOCKED : {
         MPN32 mpn = crosspage->args[0];
         retval = HostIF_IsMPNLocked(vm, mpn);
         break;
      }

      case MODULECALL_V4_LOCKPAGE: {
         void *addr = (void *)crosspage->args[0];
	 ASSERT(sizeof retval == sizeof (MPN32));
         retval = Vmx86_LockPage(vm, addr, FALSE);
	 if (PAGE_LOCK_SUCCESS(retval)) {
            if (HostIF_MarkPageDirty(addr)) {
               retval = Vmx86_UnlockPage(vm, addr);
               if (PAGE_LOCK_SUCCESS(retval)) {
                  retval = PAGE_LOCK_TOUCH_FAILED;
               }
            }
         }
         break;
      }

      case MODULECALL_V4_GET_RECYCLED_PAGE:
	 retval = Vmx86_GetRecycledPage(vm);
	 break;

      case MODULECALL_V4_SEMAWAIT:
         retval = HostIF_SemaphoreWait(vm, vcpuid, crosspage->args[0]);
	 if (retval == MX_WAITINTERRUPTED) {
	    crosspage->moduleCallInterrupted = TRUE;
	    return USERCALL_RESTART;
	 }
         break;
      case MODULECALL_V4_SEMASIGNAL:
         retval = HostIF_SemaphoreSignal(crosspage->args[0]);
	 if (retval == MX_WAITINTERRUPTED) {
	    crosspage->moduleCallInterrupted = TRUE;
	    return USERCALL_RESTART;
	 }
         break;
      case MODULECALL_V4_SEMAFORCEWAKEUP: {
         HostIF_SemaphoreForceWakeup(vm, (Vcpuid) crosspage->args[0]);
         break;
      }
      case MODULECALL_V4_IPI:
         retval = HostIF_IPI(vm, ~0);
         break;

      case MODULECALL_V4_UNLOCKPAGE: {
         void *addr = (void *)crosspage->args[0];
         Bool dirty = (Bool)crosspage->args[1];

         if (dirty) {
	    if (HostIF_MarkPageDirty(addr)) {
	       retval = PAGE_UNLOCK_TOUCH_FAILED;
	    } else {
	       retval = Vmx86_UnlockPage(vm, addr);
	    }
         } else {
	    retval = Vmx86_UnlockPage(vm, addr);
	 }
         break;
      }

      case MODULECALL_V4_RELEASE_ANON_PAGE: {
         MPN32 mpn = (MPN32)crosspage->args[0];
	 retval = Vmx86_ReleaseAnonPage(vm, mpn);
         break;
      }

      default:
         Warning("ModuleCall %d not supported\n", crosspage->moduleCallType);
      }

      /*
       * Give other threads and processes a chance.
       *
       * This is important when we are not preemptable while
       * in the driver and the monitor (e.g., Linux).
       */

      HostIF_Reschedule();
   }

   NOT_REACHED();
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetStartTime --
 *
 *      Initial the data structures that track the starting time of the
 *      virtual machine.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      none
 *
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetStartTime(VmTimeStart *st)	// OUT: return value
{
   uint32 flags;

   SAVE_FLAGS(flags);
   CLEAR_INTERRUPTS();

   st->count = RDTSC();
   st->time = HostIF_ReadUptime();

   RESTORE_FLAGS(flags);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetkHzEstimate
 *
 *      Return an estimate the of the processor's kHz rating, based on
 *      the ratio of the cycle counter and system uptime since the
 *      driver was loaded.
 *      This function could be called (on Windows) at IRQL DISPATCH_LEVEL.
 *
 * Results:
 *      Processor speed in kHz.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

uint32
Vmx86_GetkHzEstimate(VmTimeStart *st)	// IN: start time
{
   uint64 cDiff, tDiff, freq, hz;
   uint32 tmp;
   static uint32 kHz; 

   /* 
    * Cache and return the first result for consistency. 
    * TSC values can be changed without notification.
    * TSC frequency can be vary too (SpeedStep, slowing clock on HALT, etc.)
    */
   if (kHz != 0) {
      return kHz;
   }

   SAVE_FLAGS(tmp);
   CLEAR_INTERRUPTS();

   cDiff = RDTSC() - st->count;
   tDiff = HostIF_ReadUptime() - st->time;

   RESTORE_FLAGS(tmp);

   if (tDiff == 0) {
      return 0;  /* Failure */
   }

   /*
    * Compute the CPU speed in kHz, which is cDiff / (tDiff /
    * HostIF_UptimeFrequency()) / 1000.  We need to do the computation
    * carefully to avoid overflow or undue loss of precision.  Also,
    * on Linux we can't do a 64/64=64 bit division directly, as the
    * gcc stub for that is not linked into the kernel.
    */
   freq = HostIF_UptimeFrequency();
#if defined MODULE_X86_64 || !defined linux
   while (cDiff > ((uint64) -1) / freq) {
      cDiff >>= 1;
      tDiff >>= 1;
   }
   hz = (cDiff * freq) / tDiff;
   kHz = (hz + 500) / 1000;
#else
   {
      uint32 tmpkHz;
      /* On Linux we can't do a 64/64=64 bit division, as the gcc stub
       * for that is not linked into the kernel.  We'll assume that cDiff
       * * freq fits into 64 bits and that tDiff fits into 32 bits.  This
       * is safe given the values used on Linux.
       */
      Div643264(cDiff * freq, tDiff, &hz, &tmp);
      Div643232(hz + 500, 1000, &tmpkHz, &tmp);
      kHz = tmpkHz;
   }
#endif
   return kHz;

}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_CurrentVM --
 *
 *      Return the VMDriver structure associated with
 *      the handle
 *
 *
 * Results:
 *      VMDriver * structure.NULL if not available
 *
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
VMDriver *
Vmx86_CurrentVM(void *uniqHandle)
{
  VMDriver *vm;

  HostIF_GlobalLock(2);

  vm = vmDriverList;

  while (vm) {
     if (vm->uniqHandle == uniqHandle) {
        break;
     }
     vm = vm->nextDriver;
  }

  HostIF_GlobalUnlock(2);

  return vm;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetVMForProcess  --
 *
 *      Return the VMDriver structure associated with
 *      a process
 *
 *
 * Results:
 *      VMDriver * structure.NULL if not available
 *
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

VMDriver *
Vmx86_GetVMforProcess(uintptr_t processId)
{
  VMDriver *vm;

  HostIF_GlobalLock(3);

  vm = vmDriverList;

  while (vm) {
    if (vm->processID == processId) {
       break;
    }
    vm = vm->nextDriver;
  }

  HostIF_GlobalUnlock(3);

  return vm;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetNumVMs  --
 *
 *      Return the number of VMs.
 *
 * Results:
 *      The number of VMs.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int32
Vmx86_GetNumVMs()
{
   return vmCount;
}

int32
Vmx86_GetTotalMemUsage()
{
   VMDriver *vm;
   int totalmem = 0;

   HostIF_GlobalLock(15);
   vm = vmDriverList;

   while (vm) {
      totalmem += vm->wsLimitsInfo.memsize;
      vm = vm->nextDriver;
   }
   
   HostIF_GlobalUnlock(15);
   return totalmem;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetLockedPagesLimit --
 *
 *      Set the hard limit for the number of locked pages.
 *      The limit can be set only when one VM is running or if it not set.
 *
 * Results:
 *      Returns TRUE on success and FALSE on failure to set the limit
 *
 * Side effects:
 *      Hard limit changed.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_SetLockedPagesLimit(uintptr_t limit)
{
   Bool retval = FALSE;

   HostIF_GlobalLock(4);
   if (!lockedPageLimitMonitor && 
       (lockedPageLimit == 0 || (lockedPageLimit != 0 && vmCount == 1))) {
      lockedPageLimit = limit;
      retval = TRUE;
   }
   HostIF_GlobalUnlock(4);

   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetLockedPageLimitMonitorStatus --
 *
 *      Set the status of the hard limit monitor - 1 is on, 0 is off.
 *
 * Results:
 *      Returns TRUE on success and FALSE on failure to set the limit
 *
 * Side effects:
 *      Hard limit changed.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetLockedPageLimitMonitorStatus(Bool status)
{
   lockedPageLimitMonitor = status;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_ChangeLockedPagesLimit --
 *
 *      Change the hard limit for the number of locked pages.
 *      The limit can be reduced at any time.
 *
 * Results:
 *      Returns new limit.
 *
 * Side effects:
 *      Hard limit changed.
 *
 *----------------------------------------------------------------------
 */
uintptr_t
Vmx86_ChangeLockedPagesLimit(uintptr_t limit)
{
   HostIF_GlobalLock(11);
   lockedPageLimit = limit;
   HostIF_GlobalUnlock(11);
   return lockedPageLimit;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetLockedPagesLimit --
 *
 *      Get the hard limit for the number of locked pages.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Hard limit changed.
 *
 *----------------------------------------------------------------------
 */
int32
Vmx86_GetLockedPagesLimit()
{
   return lockedPageLimit;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetUnconditionalLockedPagesLimit --
 *
 *      Set the unconditional locked pages limit.
 *      Driver-wide serialization must be ensured by the caller.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      none.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetUnconditionalLockedPagesLimit(uintptr_t pageLimit)
{
   unconditionalLockedPageLimit = pageLimit;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_LockPage --
 *
 *      Lock a page.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Number of global and per-VM locked pages increased.
 *
 *----------------------------------------------------------------------
 */

MPN
Vmx86_LockPage(VMDriver *vm,		     // IN: VMDriver
	       void* addr,		     // IN: VA of the page to lock
	       Bool allowMultipleMPNsPerVA)  // IN: allow locking many pages with the same VA
{
   MPN mpn;

   /* Atomically check and reserve locked memory */
   if (!Vmx86ReserveFreePages(vm, 1)) {
      return PAGE_LOCK_LIMIT_EXCEEDED;
   }

   HostIF_VMLock(vm);
   mpn = HostIF_LockPage(vm, addr, allowMultipleMPNsPerVA);
   HostIF_VMUnlock(vm);

   if (!PAGE_LOCK_SUCCESS(mpn)) {
      Vmx86UnreserveFreePages(vm, 1);
   }

   return mpn;
}



/*
 *----------------------------------------------------------------------
 *
 * Vmx86_UnlockPage --
 *
 *      Unlock a page.
 *
 * Results:
 *      
 *
 * Side effects:
 *      Number of global and per-VM locked pages decreased.
 *
 *----------------------------------------------------------------------
 */
int
Vmx86_UnlockPage(VMDriver *vm, void *addr)
{
   int retval;
   
   HostIF_VMLock(vm);
   retval = HostIF_UnlockPage(vm, addr);
   HostIF_VMUnlock(vm);

   if (PAGE_LOCK_SUCCESS(retval)) {
      Vmx86UnreserveFreePages(vm, 1);
   }
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_UnlockPageByMPN --
 *
 *      Unlock a page.
 *
 * Results:
 *      
 *
 * Side effects:
 *      Number of global and per-VM locked pages decreased.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_UnlockPageByMPN(VMDriver *vm, // IN: VMDriver
		      MPN mpn,	    // IN: the page to unlock
		      void* addr)   // IN: optional valid VA for this MPN
{
   int retval;

   HostIF_VMLock(vm);
   retval = HostIF_UnlockPageByMPN(vm, mpn, addr);
   HostIF_VMUnlock(vm);

   if (PAGE_LOCK_SUCCESS(retval)) {
      Vmx86UnreserveFreePages(vm, 1);
   }
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_AllocLockedPage --
 *
 *      Allocate physical locked pages from the kernel.
 *
 *      Initially the pages are not mapped to any user or kernel 
 *      address space.
 *
 * Results:
 *      non-negative value on partial/full completion: actual number of
 *      allocated MPNs. MPNs of the allocted pages are copied to 
 *      the caller's buffer.
 *
 *	negative system specific error code on error (NTSTATUS on Windows, etc.)
 *
 * Side effects:
 *      Number of global and per-VM locked pages is increased.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_AllocLockedPages(VMDriver* vm,	     // IN: VMDriver
		       MPN32* mpns,	     // OUT: buffer for allocated MPNs
		       unsigned numPages,    // IN: number of pages to allocate
		       Bool kernelMPNBuffer) // IN: is the MPN buffer in kernel or user address space?
{
   int allocatedPages;

   if (!Vmx86ReserveFreePages(vm, numPages)) {
      return PAGE_LOCK_LIMIT_EXCEEDED;
   }
   
   HostIF_VMLock(vm);
   allocatedPages = HostIF_AllocLockedPages(vm, mpns, numPages, kernelMPNBuffer);
   HostIF_VMUnlock(vm);

   if (allocatedPages < 0) {
      Vmx86UnreserveFreePages(vm, numPages);
   } else if (allocatedPages < numPages) {
      Vmx86UnreserveFreePages(vm, numPages - allocatedPages);
   }

   return allocatedPages;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_FreeLockedPages --
 *
 *      Frees physical locked pages from the kernel previosly allocated 
 *      by Vmx86_AllocLockedPages().
 *
 * Results:
 *	0 on success,
 *	non-0 system specific error code on error (NTSTATUS on Windows, etc.)
 *
 * Side effects:
 *      Number of global and per-VM locked pages is decreased.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_FreeLockedPages(VMDriver *vm,	    // IN: VM instance pointer
		      const MPN32* mpns,    // IN: user or kernel array of MPNs to free 
		      unsigned numPages,    // IN: number of pages to free
		      Bool kernelMPNBuffer) // IN: is the MPN buffer in kernel or user address space?
{
   int ret;

   HostIF_VMLock(vm);
   ret = HostIF_FreeLockedPages(vm, mpns, numPages, kernelMPNBuffer);
   HostIF_VMUnlock(vm);

   if (ret == 0) {
      Vmx86UnreserveFreePages(vm, numPages);
   }

   return ret;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetLockedPageList --
 *
 *      puts MPNs of pages that were allocated by HostIF_AllocLockedPages()
 *      into user mode buffer.
 *
 * Results:
 *	non-negative number of the MPNs in the buffer on success.
 *	negative error code on error.
 *
 * Side effects:
 *      none
 *
 *----------------------------------------------------------------------
 */

int 
Vmx86_GetLockedPageList(VMDriver* vm,          // IN: VM instance pointer
			 MPN32* mpns,          // OUT: user mode buffer for MPNs
		         unsigned int numPages) // IN: size of the buffer in MPNs 
{
   int ret;
      
   HostIF_VMLock(vm);
   ret = HostIF_GetLockedPageList(vm, mpns, numPages);
   HostIF_VMUnlock(vm);
   
   return ret;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetMemInfo --
 *
 *      Return the info about all VMs.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      VMGetMemInfoArgs is filled in.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_GetMemInfo(VMDriver *curVM,
                 int32 curVMOnly,
                 VMMemInfoArgs *outArgs,
                 int outArgsLength)
{
   VMDriver *vm;
   int i;
   int outSize;
   int wantedVMs;

   HostIF_GlobalLock(7);

   if (curVMOnly) {
      wantedVMs = 1;
   } else {
      wantedVMs = vmCount;
   }

   outSize = VM_GET_MEM_INFO_SIZE(wantedVMs);
   if (outSize > outArgsLength) {
      HostIF_GlobalUnlock(7);
      return FALSE;
   }

   outArgs->numVMs = wantedVMs;
   outArgs->numLockedPages = numLockedPages;
   outArgs->maxLockedPages = lockedPageLimit;
   outArgs->unconditionalMaxLockedPages = unconditionalLockedPageLimit;
   outArgs->callerIndex = -1;
   outArgs->globalMinAllocation = globalMinAllocation;

   if (curVM != NULL) {
      if (wantedVMs == 1) {
         outArgs->memInfo[0] = curVM->memInfo;
         outArgs->callerIndex = 0;
      } else {
         vm = vmDriverList;
         i = 0;
         outArgs->callerIndex = -1;
         while (vm != NULL && i < vmCount) {
	    if (vm == curVM) {
	       outArgs->callerIndex = i;
	    }
    	    HostIF_VMLock(vm);
	    outArgs->memInfo[i] = vm->memInfo;
    	    HostIF_VMUnlock(vm);
	    i++;

	    vm = vm->nextDriver;
         }
      }
   }

   HostIF_GlobalUnlock(7);

   return TRUE;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Vmx86_GetMemInfoCopy --
 *
 *    Return the information about all VMs by copying the data out to user
 *    memory.
 *
 *    On input, outArgs->numVMs indicates how much space has been allocated for
 *    the information. On output, it indicates how much space has been
 *    filled --hpreg
 *
 * Results:
 *    TRUE on success
 *    FALSE on failure
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

Bool
Vmx86_GetMemInfoCopy(VMDriver *curVM,           // IN
                     VMMemInfoArgs *outArgs)    // IN/OUT
{
   VMMemInfoArgs *buf;
   VMDriver *vm;
   Bool ret = FALSE;

   ASSERT(curVM);

   buf = HostIF_AllocKernelMem(VM_GET_MEM_INFO_SIZE(MAX_VMS), TRUE);
   if (!buf) {
      goto err0;
   }
   if (HostIF_CopyFromUser(buf, outArgs, VM_GET_MEM_INFO_SIZE(1))) {
      goto err1;
   }

   HostIF_GlobalLock(8);

   /* Now that we have the lock, we can read vmCount --hpreg */
   if (buf->numVMs < vmCount) {
      HostIF_GlobalUnlock(8);
      goto err1;
   }

   buf->numLockedPages = numLockedPages;
   buf->maxLockedPages = lockedPageLimit;
   buf->unconditionalMaxLockedPages = unconditionalLockedPageLimit;
   buf->globalMinAllocation = globalMinAllocation;

   for (vm = vmDriverList, buf->numVMs = 0;
        vm;
        vm = vm->nextDriver, buf->numVMs++) {
      ASSERT(buf->numVMs < vmCount);
      if (vm == curVM) {
         buf->callerIndex = buf->numVMs;
      }
      HostIF_VMLock(vm);
      buf->memInfo[buf->numVMs] = vm->memInfo;
      HostIF_VMUnlock(vm);
   }
   ASSERT(buf->numVMs == vmCount);

   HostIF_GlobalUnlock(8);

   if (!HostIF_CopyToUser(outArgs, buf, VM_GET_MEM_INFO_SIZE(buf->numVMs))) {
      ret = TRUE;
   }

err1:
   HostIF_FreeKernelMem(buf);
err0:
   return ret;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetMemInfo --
 *
 *      Set the memory management information about this VM and handles
 *      admission control. We allow vm to power on if there is room for
 *      the minimum allocation for all running vms in memory.  Note that 
 *      the hard memory limit can change dynamically in windows so we
 *      don't have guarantees due to admission control.  If memory checking
 *      is disabled, the vmx passes a non-zero valid argument to indicate
 *      that we should power on anyway.  
 *
 * Results:
 *      Returns global information about the memory state in args as well
 *      as a value indicating whether or not the virtual machine was
 *      started.
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetMemInfo(VMDriver *curVM, 
		 VMMemInfoArgs *args)
{
   HostIF_GlobalLock(9);
   args->numVMs = vmCount;
   args->numLockedPages = numLockedPages;
   args->maxLockedPages = lockedPageLimit;
   args->globalMinAllocation = globalMinAllocation;
   HostIF_VMLock(curVM);
   if ((globalMinAllocation + args->memInfo->minAllocation <= lockedPageLimit ||
        args->memInfo->valid != 0) && args->memInfo->shares > 0) {
      curVM->memInfo.shares        = args->memInfo->shares;
      curVM->memInfo.guestmemSize  = args->memInfo->guestmemSize;
      curVM->memInfo.maxAllocation = args->memInfo->maxAllocation;
      curVM->memInfo.minAllocation = args->memInfo->minAllocation;
      curVM->memInfo.usedPct       = 50;
      curVM->memInfo.valid         = TRUE;

      globalMinAllocation         += args->memInfo->minAllocation;
   }
   args->memInfo->valid = curVM->memInfo.valid;
   HostIF_VMUnlock(curVM);
   HostIF_GlobalUnlock(9);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetMemUsage --
 *
 *      Set the memory usage by this vm based on its memSample data.
 *
 * Results:
 *      Updates memory allocations.
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetMemUsage(VMDriver *curVM, 
		  uintptr_t usedPct)
{
   if (usedPct <= 100) {
      HostIF_VMLock(curVM);
      curVM->memInfo.usedPct  = usedPct;
      HostIF_VMUnlock(curVM);
    }
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_PAEEnabled --
 *
 *      Is PAE enabled?
 *
 * Results:
 *      TRUE if PAE enabled.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool
Vmx86_PAEEnabled()
{
#ifdef MODULE_X86_64
   uint64 cr4;
#else
   uint32 cr4;
#endif

   GET_CR4(cr4);
   return (cr4 & CR4_PAE) != 0;
}


/*
 *----------------------------------------------------------------------
 *
 * FreeAllResources
 *
 *     Free all the resources allocated for a vm.
 *
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Memory freed.
 *
 *----------------------------------------------------------------------
 */

static void
FreeAllResources(VMDriver *vm)
{
   if (vm) {
#ifdef USE_PERFCTRS_HOSTED
      /*
       * Can't log in PerfCtr_Release() when we're holding the lock
       * and (in any case) the VM has be dequeued.
       */
      PerfCtr_Release(vm, FALSE);
#endif

      HostIF_FreeAllResources(vm);

      HostIF_GlobalLock(10);
      globalMinAllocation -= vm->memInfo.minAllocation;
      numLockedPages -= vm->memInfo.locked;
      HostIF_GlobalUnlock(10);
      if (vm->vmhost != NULL) {
         HostIF_FreeKernelMem(vm->vmhost);
      }
      HostIF_FreeKernelMem(vm);
   }
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_AckUserCall
 *
 *      Acknowledge a user call.  The call need not have completed.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Yes.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_AckUserCall(VMDriver *vm, Vcpuid vcpuid)
{
   void *crosspage = vm->crosspage[vcpuid];
   switch (vm->vmVersion) {
      case VME_TOT:
         {
	    VMCrossPageTOT* cp = crosspage;
	    ASSERT(cp->userCallRequest != MODULECALL_USERCALL_NONE);
	    cp->userCallRequest = MODULECALL_USERCALL_NONE;
	 }
	 break;
      case VME_V4:
         {
	    VMCrossPageV4* cp = crosspage;
	    ASSERT(cp->userCallRequest != MODULECALL_USERCALL_NONE);
	    cp->userCallRequest = MODULECALL_USERCALL_NONE;
	 }
	 break;
      default:
         Warning("AckUserCall with unknown personality\n");
	 break;
   }
   HostIF_AckUserCall(vm, vcpuid);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_CompleteUserCall --
 *
 *      Take actions on completion of a user call, which may or
 *      may not have been acknowledged.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      VCPU thread continues.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_CompleteUserCall(VMDriver *vm, Vcpuid vcpuid)
{
   void *crosspage = vm->crosspage[vcpuid];
   switch (vm->vmVersion) {
      case VME_TOT:
         {
	    VMCrossPageTOT* cp = crosspage;
	    if (cp->userCallRequest != MODULECALL_USERCALL_NONE) {
	       cp->userCallRequest = MODULECALL_USERCALL_NONE;
	       HostIF_AckUserCall(vm, vcpuid);
	    }
	 }
	 break;
      case VME_V4:
         {
	    VMCrossPageV4* cp = crosspage;
	    if (cp->userCallRequest != MODULECALL_USERCALL_NONE) {
	       cp->userCallRequest = MODULECALL_USERCALL_NONE;
	       HostIF_AckUserCall(vm, vcpuid);
	    }
	 }
	 break;
      default:
         Warning("CompleteUserCall with unknown personality\n");
	 break;
   }
   HostIF_AwakenVcpu(vm, vcpuid);
}


/*
 *----------------------------------------------------------------------
 *
 * CopyMemInfoV4 --
 *
 *      Convert memory management info to VMware2/3/4 format.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      info filled in.
 *
 *----------------------------------------------------------------------
 */

static void
CopyMemInfoV4(VMMemMgmtInfoV4 *info,
	      const VMDriver *vm)
{
   info->pageFaultRate = vm->memInfoCompat.pageFaultRate;
   info->lockFailRate = vm->memInfoCompat.lockFailRate;
   info->maxLocked = vm->memInfoCompat.maxLocked;
   info->currentLocked = vm->memInfo.locked;
   info->highWaterMark = vm->memInfoCompat.highWaterMark;
   info->waitingForMemoryTime = vm->memInfoCompat.waitingForMemoryTime;
   info->poweringOn = vm->memInfoCompat.poweringOn;
   info->hasFocus = vm->memInfoCompat.hasFocus;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetMemInfoV4 --
 *
 *      Return the info about all VMs.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      VMGetMemInfoArgsV4 is filled in.
 *
 *----------------------------------------------------------------------
 */

static Bool
GetMemInfoV4(VMDriver *curVM,
             int32 curVMOnly,
             VMGetMemInfoArgsV4 *outArgs,
             int outArgsLength)
{
   VMDriver *vm;
   int i;
   int outSize;
   int wantedVMs;

   if (curVMOnly) {
      wantedVMs = 1;
   } else {
      wantedVMs = vmCount;
   }

   outSize = VM_GET_MEM_INFO_SIZE_V4(wantedVMs);
   if (outSize > outArgsLength) {
      return FALSE;
   }

   outArgs->numVMs = wantedVMs;
   outArgs->numLockedPages = numLockedPages;
   outArgs->maxLockedPages = lockedPageLimit;
   outArgs->callerIndex = -1;

   if (curVM != NULL) {
      if (wantedVMs == 1) {
         CopyMemInfoV4(outArgs->memMgmtInfo, curVM);
         outArgs->callerIndex = 0;
      } else {
         vm = vmDriverList;
         i = 0;
         outArgs->callerIndex = -1;
         while (vm != NULL && i < vmCount) {
	    if (vm == curVM) {
	       outArgs->callerIndex = i;
	    }
            CopyMemInfoV4(outArgs->memMgmtInfo + i, vm);
	    i++;

	    vm = vm->nextDriver;
         }
      }
   }

   return TRUE;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Vmx86_GetMemInfoCopyV4 --
 *
 *    Return the information about all VMs by copying the data out to user
 *    memory.
 *
 *    On input, outArgs->numVMs indicates how much space has been allocated for
 *    the information. On output, it indicates how much space has been
 *    filled --hpreg
 *
 * Results:
 *    TRUE on success
 *    FALSE on failure
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

Bool
Vmx86_GetMemInfoCopyV4(VMDriver *curVM,             // IN
                       VMGetMemInfoArgsV4 *outArgs) // IN/OUT
{
   int err;
   Bool status;
   VMGetMemInfoArgsV4 hdr;   
   VMGetMemInfoArgsV4 *tmp;
   int outSize;
   int32 curVMOnly;
   int wantedVMs;

   err = HostIF_CopyFromUser(&hdr, outArgs, sizeof(VMGetMemInfoArgsV4));
   if (err != 0) {
      return FALSE;
   }

   while (1) {
      HostIF_GlobalLock(8);
      curVMOnly = hdr.numVMs == 1;
      if (curVMOnly) {
	 wantedVMs = 1;
      } else {
	 wantedVMs = vmCount;
      }
      HostIF_GlobalUnlock(8);
      outSize = VM_GET_MEM_INFO_SIZE_V4(wantedVMs);
      tmp = HostIF_AllocKernelMem(outSize, TRUE);
      if (!tmp) {
	 return FALSE;
      }
      /* GetMemInfoV4 returns error only if buffer was too small */
      HostIF_GlobalLock(8);
      status = GetMemInfoV4(curVM, curVMOnly, tmp, outSize);
      HostIF_GlobalUnlock(8);
      if (status) {
	 break;
      }
      HostIF_FreeKernelMem(tmp);
   }
   err = HostIF_CopyToUser(outArgs, tmp, outSize);
   HostIF_FreeKernelMem(tmp);
   if (err) {
      return FALSE;
   }
   return TRUE;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetMemInfoV4 --
 *
 *      Set the memory management information about this VM.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      vm->memInfo is filled in.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetMemInfoV4(VMDriver *curVM,
                   VMSetMemInfoArgsV4 *inArgs)
{
   HostIF_GlobalLock(9);

   if (inArgs->info.pageFaultRate >= 0) {
      curVM->memInfoCompat.pageFaultRate = inArgs->info.pageFaultRate;
   }
   if (inArgs->info.lockFailRate >= 0) {
      curVM->memInfoCompat.lockFailRate = inArgs->info.lockFailRate;
   }
   if (inArgs->info.maxLocked >= 0) {
      curVM->memInfoCompat.maxLocked = inArgs->info.maxLocked;
   }
   if (inArgs->info.highWaterMark >= 0) {
      curVM->memInfoCompat.highWaterMark = inArgs->info.highWaterMark;
   }
   if (inArgs->info.waitingForMemoryTime >= 0) {
      curVM->memInfoCompat.waitingForMemoryTime = inArgs->info.waitingForMemoryTime;
   }
   if (inArgs->info.poweringOn >= 0) {
      curVM->memInfoCompat.poweringOn = inArgs->info.poweringOn;
   }
   if (inArgs->info.hasFocus >= 0) {
      curVM->memInfoCompat.hasFocus = inArgs->info.hasFocus;
   }

   HostIF_GlobalUnlock(9);
}


/*
 *----------------------------------------------------------------------
 *
 * GetVMList 
 *
 *     Return the ids for all currently running VMs. Caller must hold
 *     global VM lock.
 *      
 *
 * Results:
 *      TRUE if the caller gave us enough space for the result.
 * 
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static Bool 
GetVMList(VMGetVMListResult *list, uint32 length)
{
   VMDriver *vm;
   int i = 0;

   if (vmCount * sizeof(uint32) + sizeof(VMGetVMListResult) > length) {
      return FALSE;
   }

   list->vmCount = vmCount;

   vm = vmDriverList;

   while (vm) { 
      list->vmUIDs[i++] = vm->id;
      vm = vm->nextDriver;
   }
  
   return TRUE;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetVMListCopy
 *
 *     Return the ids for all currently running VMs by copying out each
 *     
 *      
 *
 * Results:
 *      TRUE if the caller gave us enough space for the result.
 * 
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

Bool 
Vmx86_GetVMListCopy(VMGetVMListResult *list)
{
   Bool status;
   VMGetVMListResult hdr;
   VMGetVMListResult *info;
   uint32 length;

   if (HostIF_CopyFromUser(&hdr, list, sizeof(VMGetVMListResult))) {
      return FALSE;
   }

   HostIF_GlobalLock(15);

   if (vmCount > hdr.vmCount) {
      HostIF_GlobalUnlock(15);
      return FALSE;
   }

   length = sizeof(*info) + vmCount * sizeof(uint32);
   info = HostIF_AllocKernelMem(length, TRUE);
   if (!info) {
      HostIF_GlobalUnlock(15);
      return FALSE;
   }
   status = GetVMList(info, length);
   HostIF_GlobalUnlock(15);
   if (status) {
      if (HostIF_CopyToUser(list, info, length - sizeof(uint32))) {
         status = FALSE;
      }
   }
   HostIF_FreeKernelMem(info);
   return status;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetVMInfo
 *
 *     Set the static information for this VM such as its name and port.  
 *
 * Results:
 *     None.
 * 
 * Side effects:
 *     The VMDriver structure is updated with this info.
 *
 *----------------------------------------------------------------------
 */

void 
Vmx86_SetVMInfo(VMDriver *vm, VMSetVMInfoArgs *info)
{
   memcpy(vm->name, info->name, VM_NAME_LENGTH);
   memcpy(vm->userName, info->userName, VM_USER_NAME_LENGTH);
   vm->port = info->port;
   vm->state = info->state;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetVMInfo
 *
 *     Get the static information for this VM such as its name and port.  
 *
 * Results:
 *     TRUE if a VM with this uid was found, false otherwise.
 * 
 * Side effects:
 *     None.
 *
 *----------------------------------------------------------------------
 */

Bool Vmx86_GetVMInfo(uint32 vmUID, VMGetVMInfoResult *info)
{
   Bool status = FALSE;
   VMDriver *vm;

   HostIF_GlobalLock(16);

   vm = vmDriverList;

   while (vm) { 
      if (vm->id == vmUID) {
         memcpy(info->name, vm->name, VM_NAME_LENGTH);
         memcpy(info->userName, vm->userName, VM_USER_NAME_LENGTH);
         info->port = vm->port;
         info->vmUID = vm->id;
	 info->state = vm->state;
         status = TRUE;
         break;
      }
      vm = vm->nextDriver;
   }

   HostIF_GlobalUnlock(16);

   return status;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_ReleaseAnonPage --
 *
 *      Unlock a page.
 *
 * Results:
 *      PAGE_UNLOCK_NO_ERROR
 *
 * Side effects:
 *      Number of global and per-VM locked pages is decreased.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_ReleaseAnonPage(VMDriver *vm,
		      MPN mpn)
{
   int retval;

   HostIF_VMLock(vm);
   retval = HostIF_ReleaseAnonPage(vm, mpn);
   HostIF_VMUnlock(vm);

   if (PAGE_LOCK_SUCCESS(retval)) {
      Vmx86UnreserveFreePages(vm, 1);
   }
   
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetRecycledPage --
 *
 *      gets a locked page from list of pages previosly relesed 
 *      by Vmx86_ReleaseAnonPage().
 *
 * Results:
 *	MPN of the page on success
 *      INVALID_MPN if the list is empty or lock failed. 
 *      
 *      
 *
 * Side effects:
 *      Number of global and per-VM locked pages is increased.
 *
 *----------------------------------------------------------------------
 */

MPN
Vmx86_GetRecycledPage(VMDriver *vm)
{
   VPN vpn;

   HostIF_VMLock(vm);
   vpn = HostIF_GetRecycledPage(vm);
   HostIF_VMUnlock(vm);

   if (vpn != INVALID_VPN) {
      MPN mpn = Vmx86_LockPage(vm, (void*)VPN_2_VA(vpn), FALSE);
      if (PAGE_LOCK_SUCCESS(mpn)) {
	 return mpn;
      }
      /* 
       * Return the page to recycled page list.
       * XXX It's ugly but the whole function will go away with anon apge recycling.
       */
      HostIF_VMLock(vm);
      MemTrack_RecyclePage(vm->memtracker, MemTrack_LookupVPN(vm->memtracker, vpn));
      HostIF_VMUnlock(vm);
   }
   return INVALID_MPN;
}
